#include "mesh.h"
#include <QGLWidget>
#include <unordered_map>
#include "kdtree.h"

Mesh::Mesh()
	: kdtree(NULL),
	  hull_kdtree(NULL),
	  base_kdtree(NULL)
{
	clear();
}

Mesh::~Mesh()
{
	clear();
}

void Mesh::clear()
{
	vertex_orig.clear();

	b_bbox_dirty = true;

	for(size_t i = 0 ; i < childs.size() ; ++i)
		delete childs[i];
	childs.clear();
	name.clear();
	vertex.clear();
	normal.clear();
	index.clear();
	primitive_type = 0;

	local = Mat44::identity();

	if (kdtree)
		delete kdtree;
	kdtree = NULL;

	if (hull_kdtree)
		delete hull_kdtree;
	hull_kdtree = NULL;

	if (base_kdtree)
		delete base_kdtree;
	base_kdtree = NULL;
}

void Mesh::draw()
{
	glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_DOUBLE, 0, vertex.data());
    glDisableClientState(GL_COLOR_ARRAY);
	if (!normal.empty())
	{
		glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_DOUBLE, 0, normal.data());
	}
	else
		glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glActiveTexture(GL_TEXTURE1);
    glDisable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, 0);

	glPushMatrix();
    glMultMatrixd((const GLdouble*)local.data());

    glDrawRangeElements(primitive_type, 0, vertex.size() - 1, index.size(), GL_UNSIGNED_INT, index.data());

	for(size_t i = 0 ; i < childs.size() ; ++i)
		childs[i]->draw();
	glPopMatrix();
}

void Mesh::recompute_normals()
{
	normal.clear();
	normal.resize(vertex.size());
	memset(normal.data(), 0, normal.size() * sizeof(Vec3));
	switch(primitive_type)
	{
	case GL_TRIANGLES:
		for(size_t i = 0 ; i + 2 < index.size() ; i += 3)
		{
			const u_int32_t ida = index[i];
			const u_int32_t idb = index[i + 1];
			const u_int32_t idc = index[i + 2];
			const Vec3 &a = vertex[ida];
			const Vec3 &b = vertex[idb];
			const Vec3 &c = vertex[idc];
			const Vec3 &n = (b - a) ^ (c - a);
			normal[ida] += n;
			normal[idb] += n;
			normal[idc] += n;
		}
		break;
	case GL_TRIANGLE_STRIP:
		for(size_t i = 0 ; i + 2 < index.size() ; ++i)
		{
			const u_int32_t ida = index[i];
			const u_int32_t idb = index[i + 1];
			const u_int32_t idc = index[i + 2];
			const Vec3 &a = vertex[ida];
			const Vec3 &b = vertex[idb];
			const Vec3 &c = vertex[idc];
			const Vec3 &n = (b - a) ^ (c - a);
			normal[ida] += n;
			normal[idb] += n;
			normal[idc] += n;
		}
		break;
	case GL_QUADS:
		for(size_t i = 0 ; i + 3 < index.size() ; i += 4)
		{
			const u_int32_t ida = index[i];
			const u_int32_t idb = index[i + 1];
			const u_int32_t idc = index[i + 2];
			const u_int32_t idd = index[i + 3];
			const Vec3 &a = vertex[ida];
			const Vec3 &b = vertex[idb];
			const Vec3 &c = vertex[idc];
			const Vec3 &d = vertex[idd];
			const Vec3 &n0 = (b - a) ^ (c - a);
			const Vec3 &n1 = (c - a) ^ (d - a);
			normal[ida] += n0 + n1;
			normal[idb] += n0;
			normal[idc] += n0 + n1;
			normal[idd] += n1;
		}
		break;
	}
#pragma omp parallel for
	for(std::vector<Vec3>::iterator i = normal.begin() ; i < normal.end() ; ++i)
		i->normalize();

#pragma omp parallel for
	for(size_t i = 0 ; i < childs.size() ; ++i)
		childs[i]->recompute_normals();
}

void Mesh::load(const QString &filename)
{
	clear();
	if (filename.endsWith(".off", Qt::CaseInsensitive))
	{
		loadOFF(filename);
        mergeChilds();
        mergeIdenticalVertices();
        triangulate();
		return;
	}
	if (filename.endsWith(".obj", Qt::CaseInsensitive))
	{
		loadOBJ(filename);
        mergeChilds();
        mergeIdenticalVertices();
        triangulate();
        return;
	}
	if (filename.endsWith(".ply", Qt::CaseInsensitive))
	{
		loadPLY(filename);
        mergeChilds();
        mergeIdenticalVertices();
        triangulate();
        return;
	}
	if (filename.endsWith(".3ds", Qt::CaseInsensitive))
	{
		load3DS(filename);
        mergeChilds();
        mergeIdenticalVertices();
        triangulate();
        return;
	}
	if (filename.endsWith(".stl", Qt::CaseInsensitive))
	{
		loadSTL(filename);
        mergeChilds();
        mergeIdenticalVertices();
		triangulate();
        return;
	}
}

void Mesh::setBoundingBox(const std::pair<Vec3, Vec3> &bbox)
{
	b_bbox_dirty = false;
	this->bbox = bbox;
}

std::pair<Vec3, Vec3> Mesh::getBoundingBox() const
{
	if (!b_bbox_dirty)
		return bbox;

	Vec3 m(.0f,.0f,.0f);
	Vec3 M(.0f,.0f,.0f);
	if (!vertex.empty())
		m = M = vertex.front();
	for(std::vector<Vec3>::const_iterator v = vertex.begin() ; v != vertex.end() ; ++v)
	{
		m[0] = std::min(m[0], (*v)[0]);
		m[1] = std::min(m[1], (*v)[1]);
		m[2] = std::min(m[2], (*v)[2]);
		M[0] = std::max(M[0], (*v)[0]);
		M[1] = std::max(M[1], (*v)[1]);
		M[2] = std::max(M[2], (*v)[2]);
	}

	for(size_t i = 0 ; i < childs.size() ; ++i)
	{
		const std::pair<Vec3, Vec3> &bbox = childs[i]->getBoundingBox();
		if (vertex.empty() && i == 0)
		{
			m = bbox.first;
			M = bbox.second;
		}

		for(size_t j = 0 ; j < 3 ; ++j)
		{
			m[j] = std::min(m[j], bbox.first[j]);
			M[j] = std::max(M[j], bbox.second[j]);
		}
	}
	b_bbox_dirty = false;
	return bbox = std::make_pair(m, M);
}

void Mesh::swap(Mesh *mesh)
{
	name.swap(mesh->name);
	vertex.swap(mesh->vertex);
	normal.swap(mesh->normal);
	index.swap(mesh->index);
	std::swap(primitive_type, mesh->primitive_type);
	childs.swap(mesh->childs);

	std::swap(local, mesh->local);
}

void Mesh::mergeChilds()
{
    for(size_t i = 0 ; i < childs.size() ; ++i)
    {
        Mesh *child = childs[i];
        child->mergeChilds();

        const size_t offset = vertex.size();
        const size_t first_child_index = index.size();
        vertex.insert(vertex.end(), child->vertex.begin(), child->vertex.end());
        normal.insert(normal.end(), child->normal.begin(), child->normal.end());
        index.insert(index.end(), child->index.begin(), child->index.end());
		for(size_t j = first_child_index ; j < index.size() ; ++j)
            index[j] += offset;

        delete child;
    }
    const Vec3 zero = local * Vec3(0,0,0);
    for(size_t i = 0 ; i < vertex.size() ; ++i)
    {
        vertex[i] = local * vertex[i];
        normal[i] = local * normal[i] - zero;
        normal[i].normalize();
    }
    local = Mat44::identity();
	b_bbox_dirty = true;
    childs.clear();
}

namespace std
{
    template<>
    struct hash<Vec3>
    {
        inline size_t operator()(const Vec3 &v) const
        {
            size_t h = _h(v.x());
            h = (h * 31U) ^ _h(v.y());
            h = (h * 31U) ^ _h(v.z());
            return h;
        }

        hash<Vec3::component_type> _h;
    };
}

void Mesh::mergeIdenticalVertices()
{
    for(size_t i = 0 ; i < childs.size() ; ++i)
        childs[i]->mergeIdenticalVertices();

    std::unordered_map<Vec3, u_int32_t> mapper;
    std::vector<Vec3> new_vertex;
    std::vector<u_int32_t> mapping;
    mapping.reserve(vertex.size());
    for(size_t i = 0 ; i < vertex.size() ; ++i)
    {
        const Vec3 &v = vertex[i];
        const auto it = mapper.find(v);
        if (it == mapper.end())
        {
            mapper.insert(std::make_pair(v, new_vertex.size()));
            mapping.push_back(new_vertex.size());
            new_vertex.push_back(v);
        }
        else
            mapping.push_back(it->second);
    }

    vertex.swap(new_vertex);

    for(size_t i = 0 ; i < index.size() ; ++i)
        if (index[i] < mapping.size())
            index[i] = mapping[index[i]];

    recompute_normals();
}

void Mesh::triangulate()
{
    for(size_t i = 0 ; i < childs.size() ; ++i)
        childs[i]->triangulate();

    std::vector<u_int32_t> old_index;

    switch(primitive_type)
    {
    case GL_TRIANGLES:
        break;
    case GL_QUADS:
        old_index.swap(index);
        for(size_t i = 0 ; i < old_index.size() ; i += 4)
        {
            index.push_back(old_index[i]);
            index.push_back(old_index[i + 1]);
            index.push_back(old_index[i + 2]);
            index.push_back(old_index[i + 1]);
            index.push_back(old_index[i + 2]);
            index.push_back(old_index[i + 3]);
        }
        break;
    case GL_TRIANGLE_STRIP:
        old_index.swap(index);
        for(size_t i = 2 ; i < old_index.size() ; ++i)
        {
            index.push_back(old_index[i - 2]);
            index.push_back(old_index[i - 1]);
            index.push_back(old_index[i]);
        }
        break;
    }

    recompute_normals();
}

void Mesh::buildKdTree()
{
    if (kdtree)
        delete kdtree;
#pragma omp parallel
	{
#pragma omp single
		kdtree = new KdTree(vertex.data(), (Triangle*)index.data(), (Triangle*)index.data() + index.size() / 3);
	}
}

void Mesh::transform(const Mat44 &m)
{
	if (vertex.empty())
		return;
	if (!vertex_orig.empty())
		vertex = vertex_orig;
#pragma omp parallel for
		for(auto v = vertex.begin() ; v < vertex.end() ; ++v)
			*v = m * *v;

	b_bbox_dirty = true;

	recompute_normals();
}

void Mesh::backup()
{
	vertex_orig = vertex;
}

void Mesh::invertNormals()
{
#pragma omp parallel for
	for(size_t i = 0 ; i < index.size() ; i += 3)
		std::swap(index[i + 1], index[i + 2]);
	recompute_normals();
}
